% Easy Peg solitaire puzzle
% Author: LP Class, Penny, 317 class
% Brute force

/*
Board consists of list peg(C) and hole(C) for
places C.  Possible jumps are between_(C1,C2,C3).
*/


% win(N) -- succeed if N moves from initial state reaches final state.
win(N, Staten ) :-
	init(State0),
	movesTo(N, State0, Staten),
        final(Staten).

% movesTo(N, State1, Staten) -- make N moves from State1 yielding Staten

movesTo(0,State1,State1).
movesTo(N,State1,Staten) :-
	N > 0,
	move1(State1,State2),
	N1 is N-1,
	movesTo(N1, State2, Staten).

% move1(State1, State2) -- make 1 move from State1 to reach State 2
move1(State1,State4) :-
	( between_(A,B,C) ; between_(C,B,A) ),
	swap(State1, peg(A), State2),
	swap(State2, peg(B), State3),
	swap(State3, hole(C), State4).

% swap(State1, peg(A), State2) flips peg(A) to hole(A)
% swap(State1, hole(A), State2) flips hole(A) to peg(A)
% in State1, returning State2.
swap([peg(A)|State], peg(A), [hole(A)|State]).
swap([hole(A)|State], hole(A), [peg(A)|State]).
swap([Place|State1], Place0, [Place|State2]) :-
	swap(State1, Place0, State2).

% init(State0) if State0 is the initial state
init([
  peg(10), peg(20),hole(30),
	hole(11),peg(21),hole(31),
	hole(12),peg(22),hole(32),
	         hole(23)
]).


% final(State) if State is solved (one peg)
% holes(State) if State only contains holes

holes([]).
holes([hole(_)|Xs]) :- holes(Xs).
final([peg(_)|Xs]) :- holes(Xs).
final([hole(_)|Xs]) :- final(Xs).

% between_(A,B,C) if A-B-C is a possible jump up or right
% if lower left corner is 00 and upper right corner is 66.
between_(10,20,30).
between_(11,21,31).
between_(12,22,32).
between_(10,11,12).
between_(20,21,22).
between_(21,22,23).
between_(30,31,32).
