% Lecture by Joao Martins
% Author: Joao Martins and class

% LISTS VS PREDICATES
% Operationally, both options for isNum/1 work similarly.
% The one advantage of having things in a list is that they can share variables,
% e.g. [(FirstName1, LastName1), (FirstName2, LastName1)] for two family members
% that share a last name as a variable, not a ground atom.

num(1).
num(2).
num(3).
nums([1, 2, 3]).

isNum1(X) :- num(X).
isNum2(X) :- nums(L), member(X, L).













% HOW PROLOG EXECUTES

mem(X, [H|_]) :- X = H.     % Talk about cuts eventually
mem(X, [_|T]) :- mem(X, T).

% ?- mem(3, [1, 3, 2, 3]) gives you two answers
% ?- mem(3, X) gives you infinite answers answer.

mem2(X, [_|T]) :- mem2(X, T).
mem2(X, [H|_]) :- X = H.

% Since rules are executed in the order that they show up
% in the program, mem2/2 will keep looking deeper inside
% the list until the list is empty, then applying the second rule.
% The set of answers is still the same!









% ARITHMETIC

plus(X, Y, Z) :- Z is X + Y.

% ?- plus(1, 2, 3).
% yes.
% ?- plus(1, 2, X).
% X = 3
% ?- plus(1, X, 3).
% error!

% Prolog isn't an arithmetic solver. It will unify Z with the
% result of X + Y, and will fail if that's not possible.


% Simple arithmetic for the length of a list.
ln([], 0).
ln([_|T], R) :- ln(T, B), R is B + 1.







% RECURSIVE FUNCTIONS - FACTORIAL
% factorial(X, Y) :- X is 1, X = Y. % Always unifies head of rule
                                    % Makes what's happening less obvious.
%factorial(1, 1). % Can still crash the program if not enough care is taken
                  % in the other rule.
factorial(1, 1) :- !. % Cut that makes sure that once prolog
                      % reaches the base case, it stops.
factorial(N, R) :- N > 1, N1 is N - 1, factorial(N1, R1), R is N * R1.
                 % Having N > 1 here ensures that even without a ! in the
                 % base case, factorial won't blow Prolog up.
% TRACES!
% ?- trace.    % Turns on tracing. Enter to continue, a to abort.
% ?- notrace.  % Turns off tracing.
% factorial(3, R).










% CONDITIONAL

p(X) :- always1(X),
        cond(X) -> t(X) ; f(X),
        always2(X).
        
% How to rewrite this?

p(X) :- always1(X),
        cond(X), t(X),
        always2(X).

p(X) :- always1(X),
        \+ cond(X), f(X),
        always2(X).
        
% How do we change this with cuts?
p(X) :- always1(X),
        cond(X), !, t(X),
        always2(X).

p(X) :- always1(X),
        f(X),
        always2(X).
        
% If cond(X) executes, then because of the cut, prolog will
% never try to use the second rule.







% CUTS AND EFFICIENCY AND ARGUMENTS
% http://www.learnprolognow.org/lpnpage.php?pagetype=html&pageid=lpn-htmlse44

max1(X,Y,Y):- X =< Y.
max1(X,Y,X):- X > Y.
% Inefficient because only one of the above clauses can ever succeed.
% They're mutually exclusive!
% ?- max1(1, 2, X).
% ?- max1(1, 2, 2).
% ?- max1(2, 1, 3).

% Add a green cut to the first max1/3 rule.
% Trace it to make sure that less things are executing.


max2(X,Y,Z):- X =< Y, !, Z = Y. 
max2(X,_,X).
% Looks cute, right? NOPE
% ?- max2(2, 3, 2).

% The problem: the first clause should say "no, end of story"!
% But max2(2, 3, 2) fails to unify with the head of the 1st clause...
% So prolog goes to the 2nd clause... and succeeds! Oops.

max3(X,Y,Z):- X =< Y, !, Y = Z.
max3(X,_,X).

% THIS IS A RED CUT. IF YOU REMOVE IT, YOU *WILL* CHANGE THE MEANING OF THE PROGRAM.

% This also works!


% MEMBER WITH ONLY ONE ANSWER, ASK STUDENTS
member2(X, [X|_]) :- !.
member2(X, [_|T]) :- member(X, T).








% RECITATION
% This is small subproblem from the recitation.
% Three people with three different names work in three different
% professions whose names they happen to share. But they cannot
% work the profession whose name they personally have.

profs1(L) :- L = [works(N1, P1), works(N2, P2), works(N3, P3)],
            N1 \= P1, N2 \= P2, N3 \= P3,
            N1 \= N2, N1 \= N3, N2 \= N3,
            P1 \= P2, P1 \= P3, P2 \= P3,
            NP = [baker, tailor, smith],
            member(N1, NP), member(N2, NP), member(N3, NP),
            member(P1, NP), member(P2, NP), member(P3, NP).

% The problem with the above is that \= works by trying to unify
% both sides of the operator. If that succeeds, then the operator
% fails. A variable will always unify with anything else, so a
% good rule of thumb is to make sure both sides of \= are ground!




% Fixed!
profs2(L) :- L = [works(N1, P1), works(N2, P2), works(N3, P3)],
            NP = [baker, tailor, smith],
            member(N1, NP), member(N2, NP), member(N3, NP),
            member(P1, NP), member(P2, NP), member(P3, NP),
            N1 \= P1, N2 \= P2, N3 \= P3,
            N1 \= N2, N1 \= N3, N2 \= N3,
            P1 \= P2, P1 \= P3, P2 \= P3.




% The above is a single-predicate solution to a complex problem.
% With a single rule containing all of the variables, we can
% define a lot of restrictions, and then just let prolog's
% semantics perform depth-first search over the entire state space.

% But we can also write programs that direct prolog's search procedures
% in smart ways to ensure more efficiency. We end up with predicates
% for each important concept in a problem.

% In peg solitaire, we can think about how we initialise state,
% how we define moves, valid moves, state change, winning state,
% etc. That's next.